// MIDIConverter stores standard MIDI-files as ASCII-TXT-files and vice versa
// Copyright (C) 2000 Michael Bliem      email: e9271117@student.tuwien.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


import java.io.*;
import java.util.*;

import MIDIUtil;		// Utilities for handling IO, VarLenCode, MIDI-Spec...
import MIDITrack;		// instance of this class stores one MIDI-track

class MIDI
// Each Instance stores one MIDI-File with all MIDITracks
{
	boolean RMID;							// marks MIDI-file to be an RMID-file (IFF)
	int format;								// MIDI-file format or type
	int numTracks;						// MIDI-files number of tracks
	int SMPTE;								// speed if SMPTE
	int division;							// MIDI-files TimeCode
	Vector trackVector;				// Vector for the MIDItracks

	public MIDI(FileInputStream readFile) throws MIDIException
	// inits MIDI-instance
	{
		super();
		RMID = false;
		format = 1;
		numTracks = 1;
		SMPTE = 1;
		division = 384;
		trackVector = new Vector();
		if(MIDIUtil.FILETYPE) fromMIDI(readFile);
		else fromTXT(readFile);
	}

	void fromTXT(FileInputStream readFile) throws MIDIException
	// reads all MIDITracks out of a TXT-file
	{
		String s = MIDIUtil.getNextValidToken(readFile);
		if(!s.equalsIgnoreCase("MIDITXT"))
			throw new MIDIException("No valid MIDITXT-file.\n");
		MIDIUtil.setBarOffset(0);																		// default BAROFFSET here
		MIDIUtil.setProgramOffset(1);																// default PrgOFFSET here
		s = MIDIUtil.getNextValidToken(readFile);
		while(!s.equalsIgnoreCase("Track")) {
			if(s.equalsIgnoreCase("RMID")) RMID = true;
			else if(s.equalsIgnoreCase("Format"))
				format = (new Integer(MIDIUtil.getNextValidToken(readFile))).intValue();
			else if(s.equalsIgnoreCase("NumTracks"))
				numTracks = (new Integer(MIDIUtil.getNextValidToken(readFile))).intValue();
			else if(s.equalsIgnoreCase("SMPTE"))
				SMPTE = (new Integer(MIDIUtil.getNextValidToken(readFile))).intValue();
			else if(s.equalsIgnoreCase("Division"))
				division = (new Integer(MIDIUtil.getNextValidToken(readFile))).intValue();
			else if(s.equalsIgnoreCase("ProgramOffset"))
				MIDIUtil.setProgramOffset((new Integer(MIDIUtil.getNextValidToken(readFile))).intValue());
			else if(s.equalsIgnoreCase("BarOffset"))
				MIDIUtil.setBarOffset((new Integer(MIDIUtil.getNextValidToken(readFile))).intValue());
			else {
				System.out.println("Unknown header info.");
				MIDIUtil.skipLine(readFile); }
			MIDIUtil.setSmpte(SMPTE);					// for PPQN clacualtion
			MIDIUtil.setDivision(division);		// for PPQN clacualtion
			s = MIDIUtil.getNextValidToken(readFile); }
		for(int i = 0;i < numTracks; i++) {									// MIDI-tracks
			MIDITrack track = new MIDITrack(readFile);
			if(track.valid()) trackVector.addElement(track);	// skip unknown tracks
		}
		setEndOfTrack();
	}

	void fromMIDI(FileInputStream readFile) throws MIDIException
	// reads all MIDITracks out of a MIDI-file
	{
		try {
			String fileHeaderID = MIDIUtil.getString(4,readFile);
			if (fileHeaderID.equals("RIFF")) {										// IFF compatible
				RMID = true;
				readFile.skip(4);
				if(!MIDIUtil.getString(4,readFile).equals("RMID"))
					throw new MIDIException("No valid RMID-file. Unknown header.\n");
				fileHeaderID = MIDIUtil.getString(4,readFile); }
			if (!fileHeaderID.equals("MThd"))
				throw new MIDIException("No valid MIDI-file. Header not found.\n");
			int fileHeaderSIZE = MIDIUtil.getInt(4,readFile);
			if (fileHeaderSIZE != 6) {
				throw new MIDIException("No valid MIDI-file. Unknown header.\n");}
			format = MIDIUtil.getInt(2,readFile);									// Format
			numTracks = MIDIUtil.getInt(2,readFile);							// NumTracks
			int i = MIDIUtil.getInt(1,readFile);
			if(i>=0)
				division = (i<<8) + MIDIUtil.getInt(1,readFile);		// PPQN
			else {
				SMPTE = -i;
				division = MIDIUtil.getInt(1,readFile); }						// SMPTE
			MIDIUtil.setSmpte(SMPTE);					// for PPQN clacualtion
			MIDIUtil.setDivision(division);		// for PPQN clacualtion
			for(i=0; i < numTracks; i++) {												// MIDI-tracks
				MIDITrack track = new MIDITrack(readFile);
				track.setTrackNumber(i+1);
				if(track.valid()) trackVector.addElement(track);	// skip unknown tracks
			}}
		catch (IOException m) {
			throw new MIDIException("Could not read MIDI-file.\n"); }
		setEndOfTrack();
	}

	public void toTXT(FileOutputStream writeFile) throws MIDIException
	// writes all MIDITracks to a TXT-file
	{
		if(MIDIUtil.COMMENTS) {	// writes comments on top of TXT-file
			MIDIUtil.putComment("MIDI FILE IN ASCII",writeFile);
			MIDIUtil.putCopyright(writeFile);
			MIDIUtil.putComment("HEADER",writeFile); }
		MIDIUtil.putString("          MIDITXT\n",writeFile);					// marks beginning
		if(RMID) MIDIUtil.putString("RMID\n",writeFile);							// IFF compatible
		MIDIUtil.putString("Format "+new Integer(format).toString(),writeFile);
		if(numTracks != trackVector.size()) {
			numTracks = trackVector.size();
			System.out.println("Wrong number of tracks."); }
		MIDIUtil.putString("\nNumTracks "+new Integer(numTracks).toString(),writeFile);
		if(SMPTE > 1)
			MIDIUtil.putString("\nSMPTE "+new Integer(SMPTE).toString(),writeFile);
		MIDIUtil.putString("\nDivision "+new Integer(division).toString(),writeFile);
		MIDIUtil.setSmpte(SMPTE);					// for PPQN clacualtion
		MIDIUtil.setDivision(division);		// for PPQN clacualtion
		MIDIUtil.putString("\nProgramOffset "+new Integer(MIDIUtil.PROGRAMOFFSET).toString(),writeFile);
		MIDIUtil.putString("\nBarOffset "+new Integer(MIDIUtil.BAROFFSET).toString()+"\n",writeFile);
		for(int i=0; i < trackVector.size(); i++) {
			MIDITrack t = (MIDITrack)(trackVector.elementAt(i));
			if(MIDIUtil.COMMENTS) { // writes comments on top of new track
				String bar = new Integer(MIDIUtil.BAROFFSET).toString();
				String trk = (new Integer(t.trackNumber()).toString()).concat(" ");
				if(i < 9) trk = trk.concat("-");
				MIDIUtil.putComment("TRACK"+trk+"-- 000"+bar+" -- 1",writeFile); }
			t.toTXT(writeFile);																		// Writes Tracks
		}
	}

	public void toMIDI(FileOutputStream writeFile) throws MIDIException
	// writes all MIDITracks to a MIDI-file
	{
		if(RMID) {																							// IFF header
			MIDIUtil.putString("RIFF",4,writeFile);
			MIDIUtil.putInt(size() + 4,4,writeFile);							// MIDIFile + "RMID"
			MIDIUtil.putString("RMID",4,writeFile); }
		MIDIUtil.putString("MThd",4,writeFile);									// MIDI-file header
		MIDIUtil.putInt(6,4,writeFile);
		MIDIUtil.putInt(format,2,writeFile);
		if(numTracks != trackVector.size()) {
			numTracks = trackVector.size();
			System.out.println("Different number of tracks. Header corrected."); }
		MIDIUtil.putInt(numTracks,2,writeFile);
		if(SMPTE > 1) {																					// SMPTE / PPQN
			MIDIUtil.putInt(-SMPTE,1,writeFile);
			MIDIUtil.putInt(division,1,writeFile); }
		else { MIDIUtil.putInt(division,2,writeFile); }
		for(Enumeration e = trackVector.elements(); e.hasMoreElements();) {
			MIDITrack t = (MIDITrack)(e.nextElement());
			t.toMIDI(writeFile);																	// Writes Tracks
		}
	}

	public int size() throws MIDIException
	// all bytes needed to represent as MIDI-file (not RMID-file)
	{
		int size = 14;												// MThd header
		for(Enumeration e = trackVector.elements(); e.hasMoreElements();) {
			MIDITrack track = (MIDITrack)(e.nextElement());
			size += track.size();	}						// track-Bytes
		return size;
	}

	void setEndOfTrack()
	// sets EndOfTrack to all Tracks and deletes double entries of EndOfTrack
	{
		byte[] data = {0x2f};
		for(int i = 0;i < trackVector.size();i++) {
			MIDIEvent endOfTrk = new MIDIEvent();
			endOfTrk.status = 0xff;
			endOfTrk.dataVector = new Vector();
			endOfTrk.dataVector.addElement(new Integer(0x2f));
			endOfTrk.absTime = 0;
			endOfTrk.deltaTime = 0;
			endOfTrk.valid = true;
			MIDITrack t = (MIDITrack)trackVector.elementAt(i);
			if(t.eventVector.size() > 0) {
				MIDIEvent e = (MIDIEvent)t.eventVector.elementAt(t.eventVector.size()-1);
				endOfTrk.absTime = e.absTime;
				endOfTrk.deltaTime = e.absTime;
				t.deleteEvent(-1,0xff,0x0f,data);
				if(t.eventVector.size() > 0) {
					e = (MIDIEvent)t.eventVector.elementAt(t.eventVector.size()-1);
					endOfTrk.deltaTime = endOfTrk.absTime - e.absTime;
				}}
			t.insertEvent(endOfTrk);
		}
	}

	double timeCorrect(MIDI m)
	// returns value, an events absTime must be multilied to base on same timing
	{
		if(m.SMPTE == 0) m.SMPTE = 1;
		MIDIUtil.setSmpte(m.SMPTE);
		MIDIUtil.setDivision(m.division);
		double mTicks = MIDIUtil.getTicks();
		MIDIUtil.setSmpte(this.SMPTE);
		MIDIUtil.setDivision(this.division);
		double thisTicks = MIDIUtil.getTicks();
		return thisTicks/mTicks;
	}

	public void insertMIDI(MIDI insMidi)
	// inserts MIDIfile insMidi to this MIDIfile
	{
		if(insMidi == null) return;
		int d = 0;	// counter for working output
		for(int i = 0;i < insMidi.trackVector.size();i++) {
			MIDITrack insTrk = (MIDITrack)insMidi.trackVector.elementAt(i);
			int index = posOfTrack(insTrk.trackNumber());
			if(index == -1) { trackVector.addElement(insTrk); numTracks++; } // new Track
			else {
				MIDITrack thisTrk = (MIDITrack)this.trackVector.elementAt(index);
				for(int c = 0;c < insTrk.eventVector.size();c++) {
					MIDIEvent e = (MIDIEvent)insTrk.eventVector.elementAt(c);
					e.absTime = (int)((double)e.absTime * timeCorrect(insMidi));
					thisTrk.insertEvent(e);																			// new Event
					if((MIDIFrame.mText != null) && ((++d%400) == 0)) MIDIFrame.mText.append(".");
				}	// indicates working if Windows Version
			}}
		setEndOfTrack();
	}

	public void deleteMIDI(MIDI delMidi)
	// deletes MIDIfile delMidi from this MIDIfile
	{
		if(delMidi == null) return;
		int d = 0;	// counter for working output
		double timeCorrect = (double)(SMPTE*division)/(double)(delMidi.SMPTE*delMidi.division);
		for(int i = 0;i < delMidi.trackVector.size();i++) {
			MIDITrack delTrk = (MIDITrack)delMidi.trackVector.elementAt(i);
			int index = posOfTrack(delTrk.trackNumber());
			if(index >= 0) {
				MIDITrack thisTrk = (MIDITrack)this.trackVector.elementAt(index);
				for(int c = 0;c < delTrk.eventVector.size();c++) {
					MIDIEvent e = (MIDIEvent)delTrk.eventVector.elementAt(c);
					e.absTime = (int)((double)e.absTime * timeCorrect(delMidi));
					thisTrk.deleteEvent(e,e.status & 0x0f);
					if((MIDIFrame.mText != null) && ((++d%550) == 0)) MIDIFrame.mText.append(".");
				}	// indicates working if Windows Version
				if(thisTrk.eventVector.size() == 0) deleteTrack(delTrk.trackNumber()); }
			else System.out.println("Track "+delTrk.trackNumber()+" not found for edit.\n"); }
		setEndOfTrack();
	}

	public void deleteTrack(int trk)
	// deletes Track number trk
	{
		int index = posOfTrack(trk);
		if(index >= 0) {
			trackVector.removeElementAt(index);
			numTracks --; }
		else System.out.println("Track "+trk+" not found for deleting.\n");
	}

	int posOfTrack(int trk)
	// returns position of track with number trk. -1 if not exists
	{
		int index = -1;
		for(int i = 0;i < trackVector.size();i++) {
			MIDITrack t = (MIDITrack)trackVector.elementAt(i);
			if(t.trackNumber() == trk) index = i; }
		return index;
	}

	public void deleteEvent(int trk,int absTime,int status,int chn,byte[] data)
	// delete all Events in track trk (all if -1), with TimeStamp absTime (ignore if -1),
	// with status (ignore if -1) and dataBytes data (ignore if data = null) and
	// all dataBytes that are valued -1)
	{
		for(int i = 0;i < trackVector.size();i++) {
			MIDITrack t = (MIDITrack)trackVector.elementAt(i);
			int track = t.trackNumber();
			if((trk == -1) || (trk == track)) {
				t.deleteEvent(absTime,status,chn,data);
				if(t.eventVector.size() == 0) { deleteTrack(track); i--; }
			}}
		setEndOfTrack();
	}
}
