// MIDIConverter stores standard MIDI-files as ASCII-TXT-files and vice versa
// Copyright (C) 2000 Michael Bliem      email: e9271117@student.tuwien.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


import java.io.*;
import java.util.*;

import MIDIUtil;
import MIDIEventUtil;
import MIDISystemUtil;

public class MIDIEvent
// Each Instance stores one MIDI-Event with DeltaTime, Status and DataBytes
{
	int deltaTime;			// DeltaTime of Event in clocks
	int absTime;				// absolute Time of Event in clocks
	boolean valid;			// is event valid
	public int status;	// Status Byte
	Vector dataVector;	// Vector for the Data Bytes

	public boolean valid() { return valid; }

	public MIDIEvent()
	// init-Method for an Invalid empty MIDIEvent-Instance
	{
		super();
		deltaTime = -1;
		absTime = -1;
		valid = false;
		status = -1;
		dataVector = null;
	}

	public MIDIEvent(int runningStatus,int count[],FileInputStream readFile)
		throws MIDIException
	// init-Method for an MIDIEvent-Instance
	{
		super();
		deltaTime = 0;
		absTime = 0;
		valid = true;
		status = 0;
		dataVector = new Vector();
		if(MIDIUtil.FILETYPE) fromMIDI(runningStatus,count,readFile);
		else fromTXT(count,readFile);
	}

	void fromTXT(int count[],FileInputStream readFile) throws MIDIException
	// reads MIDIEvent-data out of a TXT-file
	{
		try {
			String s = MIDIUtil.getNextValidToken(readFile);
			if((s.length() == 0) || (s.equalsIgnoreCase("Track"))) {
				count[1] = -1;
				return; }
			else {
				absTime = count[0];
				deltaTime = 0;
				while((s.length() > 4) && (s.charAt(4) == ':')) {
					absTime = MIDIUtil.getAbsTime(s);
					deltaTime = absTime - count[0];
					s = MIDIUtil.getNextValidToken(readFile); }
				if(s.equals("EOF")) {
					count[1] = -1;
					return; }
			}
			if(s.equalsIgnoreCase("NoteOff")) {
				status = 0x80; MIDIEventUtil.getNoteValue(this,readFile); }
			else if(s.equalsIgnoreCase("NoteOn")) {
				status = 0x90; MIDIEventUtil.getNoteValue(this,readFile); }
			else if(s.equalsIgnoreCase("AfterTouch")) {
				status = 0xa0; MIDIEventUtil.getNoteValue(this,readFile); }
			else if(s.equalsIgnoreCase("Control")) {
				status = 0xb0; MIDIEventUtil.getControlValue(this,readFile); }
			else if(s.equalsIgnoreCase("Program")) {
				status = 0xc0; MIDIEventUtil.getProgramValue(this,readFile); }
			else if(s.equalsIgnoreCase("Pressure")) {
				status = 0xd0; MIDIEventUtil.getPressValue(this,readFile); }
			else if(s.equalsIgnoreCase("PitchBend")) {
				status = 0xe0; MIDIEventUtil.getPitchValue(this,readFile); }
			else MIDISystemUtil.getSystemValue(s,this,readFile);
			count[0] = absTime; }
		catch(Exception m) { throw new MIDIException("Syntax error in MIDITXT-file.\n");}
	}

	void fromMIDI(int runningStatus,int count[],FileInputStream readFile)
		throws MIDIException
	// reads MIDIEvent-data out of a MIDI-file; count returns read Bytes
	{
		try {
			deltaTime = MIDIUtil.getVarLen(count,readFile);
			count[1] += deltaTime;	// count[1] increments deltaTime in each Event
			absTime = count[1];			// count[1] is valid in every EventInstance
			status = readFile.read();
			count[0]++;							// count[0] increm. all Bytes read in this Event
			int length = MIDIEventUtil.eventLength(status);
			if((status & 0x80) != 0x80) {								// handling RunningStatus
				if(runningStatus == 0) {
					throw new MIDIException("No Status-Byte found.\n"); }
				else {
					dataVector.addElement(new Integer(status));
					status = runningStatus;
					length = MIDIEventUtil.eventLength(status) - 1;
				}}
			if(length <= 2){														// handling VoiceMessages
				for(int i = 0; i < length; i++) {
					dataVector.addElement(new Integer(readFile.read()));
					count[0]++;
			  }}
			else {																			// handling SystemMessages
				if(status == 0xff) {
					dataVector.addElement(new Integer(readFile.read()));
					count[0]++; }
				int l = MIDIUtil.getVarLen(count,readFile);
				for(int i = 0; i<l; i++) {
					dataVector.addElement(new Integer(readFile.read()));
					count[0]++;
				}
			}
		}
		catch(IOException m) { throw new MIDIException("Could not read Event.\n"); }
	}

	void toTXT(int track,FileOutputStream writeFile) throws MIDIException
	// writes all MIDIEvent-data to a TXT-file
	{
		if(deltaTime > 0) {																	// timeOutput in PPQN
			double ticks = MIDIUtil.getTicks();
			int beats = (int)Math.floor((double)absTime / ticks);
			int bars = beats / MIDIUtil.NUMERATOR;
			double delta = (double)absTime - ((double)beats * ticks);
			int div = (int)Math.floor(delta);
			beats -= (bars * MIDIUtil.NUMERATOR) - 1;
			bars += MIDIUtil.BAROFFSET;
			String b = new String();
			if(bars < 10) b = b.concat("0");
			if(bars < 100) b = b.concat("0");
			if(bars < 1000) b = b.concat("0");
			b = b.concat(new Integer(bars).toString());
			String c = new Integer(beats).toString();
			String d = new String();
			if(div < 10) d = d.concat("0");
			if(div < 100) d = d.concat("0");
			d = d.concat(new Integer(div).toString());
			String t = new String("TRACK");
			t = t.concat(new Integer(track).toString() + " ");
			if(track < 10) t = t.concat("-");
			if(MIDIUtil.COMMENTS && (deltaTime > delta))			// Comment on every beat
				MIDIUtil.putComment(t + "-- " + b + " -- " + c,writeFile);
			MIDIUtil.putString(b + ":" + c + ":" + d + "\n",writeFile);
		}
		String s = new String("          ");								// EventOutput
		switch(status & 0xf0) {
			case 0x80: s = s.concat(MIDIEventUtil.getNoteText(this));    break;// NoteOff
			case 0x90: s = s.concat(MIDIEventUtil.getNoteText(this));    break;// NoteOn
			case 0xa0: s = s.concat(MIDIEventUtil.getNoteText(this));    break;// AfterTouch
			case 0xb0: s = s.concat(MIDIEventUtil.getControlText(this)); break;// ControlChange
			case 0xc0: s = s.concat(MIDIEventUtil.getProgramText(this)); break;// ProgramChange
			case 0xd0: s = s.concat(MIDIEventUtil.getPressText(this));   break;// ChannelPressure
			case 0xe0: s = s.concat(MIDIEventUtil.getPitchText(this));   break;// PitchBend
			case 0xf0: s = s.concat(MIDISystemUtil.getSystemText(this));  break;// SystemMessages
			default:   s = s.concat("InvalidStatus"); }	// never reached
		MIDIUtil.putString(s + "\n",writeFile);
	}

	void toMIDI(int runningStatus,FileOutputStream writeFile)
		throws MIDIException
	// writes all MIDIEvent-data to a TXT-file
	{
		MIDIUtil.putVarLen(deltaTime,writeFile);
		if((status != runningStatus) || (status >= 0xf0))		// handling running status
			MIDIUtil.putInt(status,1,writeFile);
		Enumeration e = dataVector.elements();
		int dataSize = dataVector.size();
		Integer i;
		if(e.hasMoreElements() && (status == 0xff)) {				// handling VLC for 0xfn
			i = (Integer)(e.nextElement());
			dataSize--;
			MIDIUtil.putInt(i.intValue(),1,writeFile); }
		if(MIDIEventUtil.eventLength(status) > 2) MIDIUtil.putVarLen(dataSize,writeFile);
		while(e.hasMoreElements()) {												// handling data-Bytes
			i = (Integer)(e.nextElement());
			MIDIUtil.putInt(i.intValue(),1,writeFile);
		}
	}

	public int size(int runningStatus) throws MIDIException
	// returns all bytes that are used to represent this event in MIDI-file
	{
		int size = 1;															// bytes for VLC
		if(deltaTime > 0x7f) size++;
		if(deltaTime > 0x3fff) size++;
		if(deltaTime > 0x1fffff) size++;
		if((status != runningStatus) || (status >= 0xf0)) size++;	// runningStatus
		int i = dataVector.size();								// data-Bytes
		size += i;
		if(MIDIEventUtil.eventLength(status) > 2) {	// VarLenCode for 0xfn
			size++;
			if(status == 0xff) i--;
			if(i > 0x7f) size++;
			if(i > 0x3fff) size++;
			if(i > 0x1fffff) size++; }
		return size;
	}
}
