// MIDIConverter stores standard MIDI-files as ASCII-TXT-files and vice versa
// Copyright (C) 2000 Michael Bliem      email: e9271117@student.tuwien.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


import java.io.*;
import MIDIUtil;

public class MIDIEventUtil
// Utilities for handling MIDIEvents
{
	// ------------------------------ Global Values ------------------------------------

	static final String[] noteNameFor = {	// note name for pitch
		" c","c#"," d","d#"," e"," f","f#"," g","g#"," a","a#"," b"," h"};

	static final String[] programNameFor =  {	// GM Instrument names for ProgNr
		"Acoustic Grand","Bright Acoustic","Electric Grand","Honky-Tonk","Electric Piano 1","Electric Piano 2","Harpsichord","Clavinet",
		"Celesta","Glockenspiel","Music Box","Vibraphone","Marimba","Xylophone","Tubular Bells","Dulcimer",
		"Drawbar Organ","Percussive Organ","Rock Organ","Church Organ","Reed Organ","Accoridan","Harmonica","Tango Accordian",
		"Nylon String Guitar","Steel String Guitar","Electric Jazz Guitar","Electric Clean Guitar","Electric Muted Guitar","Overdriven Guitar","Distortion Guitar","Guitar Harmonics",
		"Acoustic Bass","Electric Bass (finger)","Electric Bass (pick)","Fretless Bass","Slap Bass 1","Slap Bass 2","Synth Bass 1","Synth Bass 2",
		"Violin","Viola","Cello","Contrabass","Tremolo Strings","Pizzicato Strings","Orchestral Strings","Timpani",
		"String Ensemble 1","String Ensemble 2","SynthStrings 1","SynthStrings 2","Choir Aahs","Voice Oohs","Synth Voice","Orchestra Hit",
		"Trumpet","Trombone","Tuba","Muted Trumpet","French Horn","Brass Section","SynthBrass 1","SynthBrass 2",
		"Soprano Sax","Alto Sax","Tenor Sax","Baritone Sax","Oboe","English Horn","Bassoon","Clarinet",
		"Piccolo","Flute","Recorder","Pan Flute","Blown Bottle","Skakuhachi","Whistle","Ocarina",
		"Lead 1 (square)","Lead 2 (sawtooth)","Lead 3 (calliope)","Lead 4 (chiff)","Lead 5 (charang)","Lead 6 (voice)","Lead 7 (fifths)","Lead 8 (bass+lead)",
		"Pad 1 (new age)","Pad 2 (warm)","Pad 3 (polysynth)","Pad 4 (choir)","Pad 5 (bowed)","Pad 6 (metallic)","Pad 7 (halo)","Pad 8 (sweep)",
		"FX 1 (rain)","FX 2 (soundtrack)","FX 3 (crystal)","FX 4 (atmosphere)","FX 5 (brightness)","FX 6 (goblins)","FX 7 (echoes)","FX 8 (sci-fi)",
		"Sitar","Banjo","Shamisen","Koto","Kalimba","Bagpipe","Fiddle","Shanai",
		"Tinkle Bell","Agogo","Steel Drums","Woodblock","Taiko Drum","Melodic Tom","Synth Drum","Reverse Cymbal",
		"Guitar Fret Noise","Breath Noise","Seashore","Bird Tweet","Telephone Ring","Helicopter","Applause","Gunshot"};

	static final String[] controlNameFor =  {	// Controller names for CtrlNr
		"Bank Select (coarse)","Modulation Wheel (coarse)","Breath controller (coarse)","","Foot Pedal (coarse)","Portamento Time (coarse)","Data Entry (coarse)","Volume (coarse)",
		"Balance (coarse)","","Pan position (coarse)","Expression (coarse)","Effect Control 1 (coarse)","Effect Control 2 (coarse)","","",
		"General Purpose Slider 1","General Purpose Slider 2","General Purpose Slider 3","General Purpose Slider 4","","","","",
		"","","","","","","","",
		"Bank Select (fine)","Modulation Wheel (fine)","Breath controller (fine)","","Foot Pedal (fine)","Portamento Time (fine)","Data Entry (fine)","Volume (fine)",
		"Balance (fine)","","Pan position (fine)","Expression (fine)","Effect Control 1 (fine)","Effect Control 2 (fine)","","",
		"","","","","","","","",
		"","","","","","","","",
		"Hold Pedal (on/off)","Portamento (on/off)","Sustenuto Pedal (on/off)","Soft Pedal (on/off)","Legato Pedal (on/off)","Hold 2 Pedal (on/off)","Sound Variation","Sound Timbre",
		"Sound Release Time","Sound Attack Time","Sound Brightness","Sound Control 6","Sound Control 7","Sound Control 8","Sound Control 9","Sound Control 10",
		"General Purpose Button 1 (on/off)","General Purpose Button 2 (on/off)","General Purpose Button 3 (on/off)","General Purpose Button 4 (on/off)","","","","",
		"","","","Effects Level","Tremulo Level","Chorus Level","Celeste Level","Phaser Level",
		"Data Button increment","Data Button decrement","Non-registered Parameter (fine)","Non-registered Parameter (coarse)","Registered Parameter (fine)","Registered Parameter (coarse)","","",
		"","","","","","","","",
		"","","","","","","","",
		"All Sound Off","All Controllers Off","Local Keyboard (on/off)","All Notes Off","Omni Mode Off","Omni Mode On","Mono Operation","Poly Operation"};


	// ----------------------------- MIDI Spec Methods ---------------------------------


	public static int eventLength(int status) throws MIDIException
	// returns how many data bytes follow this status. 3 if complex (SYSEX).
	{
		int i = 2;
		if((status >= 0xc0) && (status <= 0xdf)) i = 1;
		if(status >= 0xf0)
			if((status == 0xf0) || (status == 0xf7) || (status == 0xff)) i = 3;
			else throw new MIDIException("Illegal Status Byte.\n");
		return i;
	}

	static String channelToText(int status)
	// returns String (" Ch_n") with Channel number n
	{
		String s = new String();
		if (status < 0xf0) {
			s = s.concat(" Ch");
			s = s.concat(new Integer((status & 0x0f)+1).toString());
			if((status & 0x0f) < 9) s = s.concat(" "); }
		return s;
	}

	static int textToChannel(String s) throws MIDIException
	// returns int with channel number stored in String s
	{
		try { return new Integer(s.substring(2)).intValue() - 1; }
		catch(Exception m) { throw new MIDIException("Syntax error in MIDITXT-file.\n"); }
	}

	static int textToNote(String s) throws MIDIException
	// returns int with noteValue stored in String s
	{
		try {
			int len = 1;
			int note = 0;
			String n = new String();
			if((s.charAt(1) == '#') || (s.charAt(1) == 'b')) len = 2;
			else n = n.concat(" ");
			n = n.concat(s.substring(0,len));
			for(int i = 0; i <= 12; i++) if(n.equalsIgnoreCase(noteNameFor[i])) note = i;
			if(note == 12) note--;
			return note + (new Integer(s.substring(len)).intValue() + 2) * 12; }
		catch(Exception m) { throw new MIDIException("Syntax error in MIDITXT-file.\n"); }
	}

	static String valueToText(int pos,MIDIEvent e) throws MIDIException
	// returns String of dataByteValue number pos of event e
	{
		String s = new String(" ");
		if((e.dataVector.size() <= pos) || (pos < 0))
			throw new MIDIException("Invalid dataByte-index.\n");
		int i = ((Integer)(e.dataVector.elementAt(pos))).intValue();
		if(i < 10) s = s.concat(" ");
		if(i < 100) s = s.concat(" ");
		return s.concat(new Integer(i).toString());
	}


	// ----------------------------- MIDI Event Methods --------------------------------


	static String getNoteText(MIDIEvent e) throws MIDIException
	// returns String with NoteOn, NoteOff or AfterTouch event
	{
		String s = new String();
		if(e.dataVector.size() != 2) throw new MIDIException("Invalid NoteMessage.\n");
		switch(e.status & 0xf0) {
			case 0x80: s = s.concat("NoteOff   "); break;
			case 0x90: s = s.concat("NoteOn    "); break;
			case 0xa0: s = s.concat("AfterTouch"); break;
			default: throw new MIDIException("Wrong status.\n"); } // never reached
		s = s.concat(channelToText(e.status)+" ");
		Integer i = (Integer)(e.dataVector.elementAt(0));
		int note = i.intValue() % 12;
		if((note == 11) && MIDIUtil.GERMAN) note++;
		s = s.concat(noteNameFor[note]);
		int octave = (i.intValue() / 12) - 2;
		s = s.concat(new Integer(octave).toString());
		if(octave >= 0) s = s.concat(" ");
		return s.concat(valueToText(1,e));
	}

	static void getNoteValue(MIDIEvent e,FileInputStream fpp) throws MIDIException
	// reads dataBytes of Note and Touch out of TXT-file and stores in event e
	{
		String s = MIDIUtil.getNextValidToken(fpp);
		e.status += textToChannel(s);
		s = MIDIUtil.getNextValidToken(fpp);
		e.dataVector.addElement(new Integer(textToNote(s)));
		e.dataVector.addElement(new Integer(MIDIUtil.getNextValidToken(fpp)));
	}

	static String getControlText(MIDIEvent e) throws MIDIException
	// returns String with ControlChange event and ControllerName if COMMENTS
	{
		String s = new String("Control   ");
		if(e.dataVector.size() != 2) throw new MIDIException("Invalid ControlMessage.\n");
		s = s.concat(channelToText(e.status));
		int controller = ((Integer)(e.dataVector.elementAt(0))).intValue();
		s = s.concat(valueToText(0,e) + " " + valueToText(1,e));
		if(MIDIUtil.COMMENTS) s = s.concat("   // " + controlNameFor[controller]);
		return s;
	}

	static void getControlValue(MIDIEvent e,FileInputStream fpp) throws MIDIException
	// reads dataBytes of Control out of TXT-file and stores in event e
	{
		String s = MIDIUtil.getNextValidToken(fpp);
		e.status += textToChannel(s);
		e.dataVector.addElement(new Integer(MIDIUtil.getNextValidToken(fpp)));
		e.dataVector.addElement(new Integer(MIDIUtil.getNextValidToken(fpp)));
	}

	static String getProgramText(MIDIEvent e) throws MIDIException
	// returns String with ProgramChange event and InstrumentName if COMMENTS
	{
		String s = new String("Program   ");
		if(e.dataVector.size() != 1) throw new MIDIException("Invalid ProgramMessage.\n");
		s = s.concat(channelToText(e.status) + " ");
		int i = ((Integer)(e.dataVector.elementAt(0))).intValue();
		if(i < 0) i = 0; if(i > 127) i = 127;
		i += MIDIUtil.PROGRAMOFFSET;
		if(i < 10) s = s.concat(" ");
		if(i < 100) s = s.concat(" ");
		s = s.concat(new Integer(i).toString());
		if(MIDIUtil.COMMENTS)
			s = s.concat("        // GM: " + programNameFor[i - MIDIUtil.PROGRAMOFFSET]);
		return s;
	}

	static void getProgramValue(MIDIEvent e,FileInputStream fpp) throws MIDIException
	// reads dataBytes of Program out of TXT-file and stores in event e
	{
		String s = MIDIUtil.getNextValidToken(fpp);
		e.status += textToChannel(s);
		int i = new Integer(MIDIUtil.getNextValidToken(fpp)).intValue()-MIDIUtil.PROGRAMOFFSET;
		if(i < 0) i = 0; if(i > 127) i = 127;
		e.dataVector.addElement(new Integer(i));
	}

	static String getPressText(MIDIEvent e) throws MIDIException
	// returns String with ChannelPressure event
	{
		String s = new String("Pressure  ");
		if(e.dataVector.size() != 1) throw new MIDIException("Invalid PressureMessage.\n");
		return s.concat(channelToText(e.status)+"     " + valueToText(0,e));
	}

	static void getPressValue(MIDIEvent e,FileInputStream fpp) throws MIDIException
	// reads dataBytes of Pressure out of TXT-file and stores in event e
	{
		String s = MIDIUtil.getNextValidToken(fpp);
		e.status += textToChannel(s);
		e.dataVector.addElement(new Integer(MIDIUtil.getNextValidToken(fpp)));
	}

	static String getPitchText(MIDIEvent e) throws MIDIException
	// returns String with PitchBendEvent. value as ("'+'/'-'8132")
	{
		String s = new String("PitchBend ");
		if(e.dataVector.size() != 2)
			throw new MIDIException("Invalid PitchBendMessage.\n");
		s = s.concat(channelToText(e.status));
		int first =  ((Integer)(e.dataVector.elementAt(0))).intValue();
		int second = ((Integer)(e.dataVector.elementAt(1))).intValue();
		int i = ((second << 7) | first) - 0x2000;
		s = s.concat("    ");
		if(Math.abs(i) < 10) s = s.concat(" ");
		if(Math.abs(i) < 100) s = s.concat(" ");
		if(Math.abs(i) < 1000) s = s.concat(" ");
		if(i > 0) s = s.concat("+");
		if(i ==0) s = s.concat(" ");
		return s.concat(new Integer(i).toString());
	}

	static void getPitchValue(MIDIEvent e,FileInputStream fpp) throws MIDIException
	// reads dataBytes of Pressure out of TXT-file and stores in event e
	{
		String s = MIDIUtil.getNextValidToken(fpp);
		e.status += textToChannel(s);
		s = MIDIUtil.getNextValidToken(fpp);
		if(s.startsWith("+")) s = s.substring(1);
		int value = new Integer(s).intValue() + 0x2000;
		e.dataVector.addElement(new Integer(value & 0x7f));
		e.dataVector.addElement(new Integer(value >> 7));
	}
}