// MIDIConverter stores standard MIDI-files as ASCII-TXT-files and vice versa
// Copyright (C) 2000 Michael Bliem      email: e9271117@student.tuwien.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


import java.io.*;
import MIDIUtil;
import MIDIEventUtil;

public class MIDISystemUtil
// Utilities for handling MIDISystemEvents
{
	static final char[] hexCharFor = {	// Hex names for integers
		'0','1','2','3','4','5','6','7','8','9','a','b','c','d','e','f'};

	static final String[] manufactorNameFor =  {	// Manufactor names for first DataByte
		"","Sequential Circuits","Big Briar","Octave / Plateau","Moog","Passport Designs","Lexicon","Kurzweil",
		"Fender","Gulbransen","Delta Labs","Sound Comp.","General Electro","Techmar","Matthews Research","",
		"Oberheim","PAIA","Simmons","DigiDesign","Fairlight","JL Cooper","Lowery","Lin",
		"Emu","","","Peavey","","","","",
		"Bon Tempi","S.I.E.L.","","SyntheAxe","Hohner","Crumar","Solton","Jellinghaus Ms",
		"CTS","PPG","","","","","","Elka",
		"","","","","","","Cheetah","",
		"","","","","","","Waldorf","",
		"Kawai","Roland","Korg","Yamaha","Casio","Akai","",""};

	// -------------------------- SYSEX Handling Methods -------------------------------


	static String getText(MIDIEvent e) throws MIDIException
  // returns String of Text in MetaEvent e
	{
		byte b;
		String s = new String(" \"");
		if(e.dataVector.size() < 1)
			throw new MIDIException("Invalid TextSearch in MetaEvent.\n");
		byte byteBuffer[] = new byte[e.dataVector.size()-1];
		for(int i = 1; i < e.dataVector.size(); i++) {
			byteBuffer[i-1] = (byte)(((Integer)(e.dataVector.elementAt(i))).intValue());
			if(byteBuffer[i-1] == 0) byteBuffer[i-1] = 0x20; }	// Space instead of \0
		return s.concat(new String(byteBuffer) + "\"");
	}

	static void storeText(MIDIEvent e,FileInputStream fpp) throws MIDIException
	// reads text from fpp and stores in dataBytes of event e
	{
		String s = MIDIUtil.getNextValidToken(fpp);
		if((s.charAt(0) != '\"') || (s.charAt(s.length()-1) != '\"'))
			throw new MIDIException("Syntax error: unqouted text.\n");
		byte[] byteBuffer = s.getBytes();
		for(int i = 1; i < (s.length()-1); i++)
			e.dataVector.addElement(new Integer((int)byteBuffer[i]));
	}

	static String getHexText(int first,int last,MIDIEvent e) throws MIDIException
	// returns String with Event e dataBytes (from first to last) in Hex
	{
		String s = new String();
		if((e.dataVector.size() <= last) || (first > last) || (first < 0) || (last < 0))
			throw new MIDIException("Wrong number of data-Bytes.\n");
		for(int i = first; i <= last; i++) {
			Integer d = (Integer)(e.dataVector.elementAt(i));
			s = s.concat(" ");
			if(d.intValue() < 16) s = s.concat("0");
			s = s.concat(Integer.toHexString(d.intValue())); }
		return s.concat(" END");
	}

	static void getHexValue(MIDIEvent e,FileInputStream fpp) throws MIDIException
	// reads HexValues out of fpp and writes into Events e DataBytes
	{
		int hex = 0;
		String s = MIDIUtil.getNextValidToken(fpp);
		if(s.equals("f0") || s.equals("f7")) s = MIDIUtil.getNextValidToken(fpp);
		while(!s.equalsIgnoreCase("END") && !s.equals("EOF")) {
			for(int i = 0; i<16; i++) if(s.charAt(0) == hexCharFor[i]) hex = i << 4;
			for(int i = 0; i<16; i++) if(s.charAt(1) == hexCharFor[i]) hex += i;
			e.dataVector.addElement(new Integer(hex));
			s =  MIDIUtil.getNextValidToken(fpp); }
		if(s.equals("EOF")) throw new MIDIException("Syntax error: \'END\' not found.\n");
	}

	static String sysexNameFor(MIDIEvent e)
	// returns String with name of SysEx-message if known
	{
		int i;
		String s = new String();
		if(e.dataVector.size() == 5) {										// testing for GM System on/off
			if((((Integer)(e.dataVector.elementAt(0))).intValue() == 0x7e) &&
				 (((Integer)(e.dataVector.elementAt(1))).intValue() == 0x7f) &&
				 (((Integer)(e.dataVector.elementAt(2))).intValue() == 0x09) &&
				 (((Integer)(e.dataVector.elementAt(4))).intValue() == 0xf7)) {
				if(((Integer)(e.dataVector.elementAt(3))).intValue() == 0x00)
					s = s.concat("GM System Disabled");
				if(((Integer)(e.dataVector.elementAt(3))).intValue() == 0x01)
					s = s.concat("GM System Enabled");
			}}
		if(e.dataVector.size() == 7) {												// testing for Master Volume
			if((((Integer)(e.dataVector.elementAt(0))).intValue() == 0x7f) &&
				 (((Integer)(e.dataVector.elementAt(1))).intValue() == 0x7f) &&
				 (((Integer)(e.dataVector.elementAt(2))).intValue() == 0x04) &&
				 (((Integer)(e.dataVector.elementAt(3))).intValue() == 0x01) &&
				 (((Integer)(e.dataVector.elementAt(6))).intValue() == 0xf7)) {
				i = ((Integer)(e.dataVector.elementAt(5))).intValue() << 7;
				i += ((Integer)(e.dataVector.elementAt(4))).intValue();
				s = s.concat("Master Volume " + new Integer(i).toString());
			}}
		if((s.length() == 0) && (e.dataVector.size() > 0)) {
			i = ((Integer)(e.dataVector.elementAt(0))).intValue();
			if(i < 0x48) s = s.concat(manufactorNameFor[i]);						// Manufactor ID
			if(i == 0x7e) s = s.concat("Non-RealTime");
			if(i == 0x7f) s = s.concat("RealTime");
			if((i >= 0x7e) && (e.dataVector.size() > 1)) {			// MIDI Sample Dump Standard
				int c = ((Integer)(e.dataVector.elementAt(1))).intValue();
				if(c < 0x7f) s = s.concat(" at channel " + new Integer(c).toString());
				if(c ==0x7f) s = s.concat(" disregard channel");
			}}
		return s;
	}

	static String getEscapedText(MIDIEvent e) throws MIDIException
	// returns String with escaped SystemMessage
	{
		int i;
		String s = new String();
		if(e.dataVector.size() < 1)
			throw new MIDIException("Invalid EscapedSystemMessage.\n");
		switch(((Integer)(e.dataVector.elementAt(0))).intValue()) {
			case 0xf1:
				s = s.concat("MTC_QTMessage       " + MIDIEventUtil.valueToText(1,e)); break;
			case 0xf2: {
				s = s.concat("SongPosPointer     ");
				if(e.dataVector.size() != 3)
					throw new MIDIException("Invalid SongPosMessage.\n");
				int first =  ((Integer)(e.dataVector.elementAt(1))).intValue();
				int second = ((Integer)(e.dataVector.elementAt(2))).intValue();
				i = ((second << 7) | first);
				if(i < 10) s = s.concat(" ");
				if(i < 100) s = s.concat(" ");
				if(i < 1000) s = s.concat(" ");
				if(i < 10000) s = s.concat(" ");
				s = s.concat(new Integer(i).toString());
			} break;
			case 0xf3:
				s = s.concat("SongSelect          " + MIDIEventUtil.valueToText(1,e)); break;
			case 0xf6: s = s.concat("TuneRequest"); break;
			case 0xf8: s = s.concat("MIDI_Clock"); break;
			case 0xf9: s = s.concat("Tick"); break;
			case 0xfA: s = s.concat("MIDI_Start"); break;
			case 0xfB: s = s.concat("MIDI_Continue"); break;
			case 0xfC: s = s.concat("MIDI_Stop"); break;
			case 0xfE: s = s.concat("ActiveSense"); break;
			default:   s = s.concat("UnknownEscaped");	// future features
		}
		return s;
	}

	static String getMetaEventText(MIDIEvent e) throws MIDIException
	// returns String with MetaEvent
	{
		int i;
		String s = new String();
		if(e.dataVector.size() < 1)
			throw new MIDIException("Invalid MetaEventMessage.\n");
		switch(((Integer)(e.dataVector.elementAt(0))).intValue()) {
			case 0x00: {																								// Sequenze Number
					s = s.concat("SequenzeNumber     ");
					if(e.dataVector.size() != 3)
						throw new MIDIException("Invalid SequenceNumberMessage.\n");
					i =  ((Integer)(e.dataVector.elementAt(1))).intValue() << 8;
					i += ((Integer)(e.dataVector.elementAt(2))).intValue();
					if(i < 10) s = s.concat(" ");
					if(i < 100) s = s.concat(" ");
					if(i < 1000) s = s.concat(" ");
					if(i < 10000) s = s.concat(" ");
					s = s.concat(new Integer(i).toString());
				} break;
			case 0x01: s = s.concat("Text           " + getText(e)); break;	// Text
			case 0x02: s = s.concat("Copyright      " + getText(e)); break;	// Copyright
			case 0x03: s = s.concat("SequenzeName   " + getText(e)); break;	// SeqName
			case 0x04: s = s.concat("Instrument     " + getText(e)); break;	// Instrument
			case 0x05: s = s.concat("Lyric          " + getText(e)); break;	// Lyric
			case 0x06: s = s.concat("Marker         " + getText(e)); break;	// Marker
			case 0x07: s = s.concat("CuePoint       " + getText(e)); break;	// CuePoint
			case 0x20: s = s.concat("MIDI_Channel   " + MIDIEventUtil.valueToText(1,e)); break;
			case 0x21: s = s.concat("MIDI_Port      " + MIDIEventUtil.valueToText(1,e)); break;
			case 0x2f: s = s.concat("EndOfTrack     "); break;							// EndOfTrack
			case 0x51: {																										// Tempo
				s = s.concat("Tempo           ");
				if(e.dataVector.size() != 4)
					throw new MIDIException("Invalid TempoMessage.\n");
				int t1 = ((Integer)(e.dataVector.elementAt(1))).intValue();
				int t2 = ((Integer)(e.dataVector.elementAt(2))).intValue();
				int t3 = ((Integer)(e.dataVector.elementAt(3))).intValue();
				i = (((t1 << 8) + t2) << 8) + t3;
				MIDIUtil.setTempo(i);	// for PPQN calculation
				s = s.concat(new Integer(i).toString());
				if(MIDIUtil.COMMENTS) s = s.concat("   // " + (60000000.0 /(double)i) + " BPM");
				} break;
			case 0x54: {																										// SMPTE Offset
				s = s.concat("SMPTE_Offset    ");
				if(e.dataVector.size() != 6)
					throw new MIDIException("Invalid SMPTE_OffsetMessage.\n");
				int hr = ((Integer)(e.dataVector.elementAt(1))).intValue();
				int mn = ((Integer)(e.dataVector.elementAt(2))).intValue();
				int se = ((Integer)(e.dataVector.elementAt(3))).intValue();
				int fr = ((Integer)(e.dataVector.elementAt(4))).intValue();
				int ff = ((Integer)(e.dataVector.elementAt(5))).intValue();
				if(hr < 10) s = s.concat("0");
				s = s.concat(new Integer(hr).toString() + ":");
				if(mn < 10) s = s.concat("0");
				s = s.concat(new Integer(mn).toString() + ":");
				if(se < 10) s = s.concat("0");
				s = s.concat(new Integer(se).toString() + ":");
				if(fr < 10) s = s.concat("0");
				s = s.concat(new Integer(fr).toString() + ":");
				if(ff < 10) s = s.concat("0");
				s = s.concat(new Integer(ff).toString());
				} break;
			case 0x58: {																									// Time Signature
				s = s.concat("TimeSignature   ");
				if(e.dataVector.size() != 5)
					throw new MIDIException("Invalid TimeSignatureMessage.\n");
				int num = ((Integer)(e.dataVector.elementAt(1))).intValue();
				int den = ((Integer)(e.dataVector.elementAt(2))).intValue();
				int met = ((Integer)(e.dataVector.elementAt(3))).intValue();
				int mqn = ((Integer)(e.dataVector.elementAt(4))).intValue();
				MIDIUtil.setNumerator(num);		// for PPQN clacualtion
				MIDIUtil.setDenominator(den);	// for PPQN clacualtion
				s = s.concat(new Integer(num).toString() + "/");
				s = s.concat(new Integer(1 << den).toString() + " ");
				s = s.concat(new Integer(met).toString() + " ");
				s = s.concat(new Integer(mqn).toString());
				} break;
			case 0x59: {																									// Key Signature
				s = s.concat("KeySignature    ");
				if(e.dataVector.size() != 3)
					throw new MIDIException("Invalid KeySignatureMessage.\n");
				int sf = ((Integer)(e.dataVector.elementAt(1))).intValue();
				int mi = ((Integer)(e.dataVector.elementAt(2))).intValue();
				if(Math.abs(sf) < 10) s = s.concat(" ");
				if(Math.abs(sf) < 100) s = s.concat(" ");
				s = s.concat(new Integer(Math.abs(sf)).toString()+" ");
				if(sf < 0) s = s.concat("flat ");
				if(sf > 0) s = s.concat("sharp");
				if(sf ==0) s = s.concat("key_c");
				if(mi ==0) s = s.concat(" major");
				if(mi ==1) s = s.concat(" minor");
				} break;
			case 0x7f:																									// Proprietary Event
				s = s.concat("ProprietaryEvnt" + getHexText(1,e.dataVector.size()-1,e));
				break;
			default:   s = s.concat("UnknownMetaEvnt");	// future features
		}
		return s;
	}

	static String getSystemText(MIDIEvent e) throws MIDIException
	// returns Text of whole System Message
	{
		int i;
		String s = new String();
		if(e.dataVector.size() > 0)																// get first dataByte
			i = ((Integer)(e.dataVector.elementAt(0))).intValue();
		else throw new MIDIException("MIDI-event is not valid.\n");
		switch(e.status) {
			case 0xf0: {																					// System Exclusive
				s = s.concat("SystemExclusive f0");
				s = s.concat(getHexText(0,e.dataVector.size()-1,e));
				if(MIDIUtil.COMMENTS) s = s.concat("   // " + sysexNameFor(e));
				} break;
			case 0xf7:
				if(i < 0xf0) {																			// SysExContinue
					s = s.concat("SysExContinue   f7");
					s = s.concat(getHexText(0,e.dataVector.size()-1,e)); }
				else s = s.concat(getEscapedText(e)); 							// EscapedSystem
				break;
			case 0xff: s = s.concat(getMetaEventText(e)); break;	// Meta Events
			default: s = s.concat("UnknownStatus"); }	// future feature
		return s;
	}

	static void getSystemValue(String s,MIDIEvent e,FileInputStream fpp)
		throws MIDIException
	// reads dataBytes of System out of TXT-file and stores in event e
	{
		int v = 0;
		Integer i;
		if(s.equalsIgnoreCase("SystemExclusive")) {											// SystemExclusive
			e.status = 0xf0; getHexValue(e,fpp); }
		else if(s.equalsIgnoreCase("SysExContinue")) {									// SysExContinue
			e.status = 0xf7; getHexValue(e,fpp); }
		else if(s.equalsIgnoreCase("MTC_QTMessage")) {									// MTC QTMessage
			e.status = 0xf7; e.dataVector.addElement(new Integer(0xf1));
			i = new Integer(MIDIUtil.getNextValidToken(fpp));
			e.dataVector.addElement(i); }
		else if(s.equalsIgnoreCase("SongPosPointer")) {									// SongPosPointer
			e.status = 0xf7; e.dataVector.addElement(new Integer(0xf2));
			v = new Integer(MIDIUtil.getNextValidToken(fpp)).intValue();
			e.dataVector.addElement(new Integer(v & 0x7f));
			e.dataVector.addElement(new Integer(v >> 7)); }
		else if(s.equalsIgnoreCase("SongSelect")) {											// Song Select
			e.status = 0xf7; e.dataVector.addElement(new Integer(0xf3));
			i = new Integer(MIDIUtil.getNextValidToken(fpp));
			e.dataVector.addElement(i); }
		else if(s.equalsIgnoreCase("TuneRequest")) {										// Tune Request
			e.status = 0xf7; e.dataVector.addElement(new Integer(0xf6)); }
		else if(s.equalsIgnoreCase("MIDI_Clock")) {											// MIDI Clock
			e.status = 0xf7; e.dataVector.addElement(new Integer(0xf8)); }
		else if(s.equalsIgnoreCase("Tick")) {														// Tick
			e.status = 0xf7; e.dataVector.addElement(new Integer(0xf9)); }
		else if(s.equalsIgnoreCase("MIDI_Start")) {											// MIDI Start
			e.status = 0xf7; e.dataVector.addElement(new Integer(0xfa)); }
		else if(s.equalsIgnoreCase("MIDI_Continue")) {									// MIDI Continue
			e.status = 0xf7; e.dataVector.addElement(new Integer(0xfb)); }
		else if(s.equalsIgnoreCase("MIDI_Stop")) {											// MIDI Stop
			e.status = 0xf7; e.dataVector.addElement(new Integer(0xfc)); }
		else if(s.equalsIgnoreCase("ActiveSense")) {										// Active Sense
			e.status = 0xf7; e.dataVector.addElement(new Integer(0xfe)); }
		else if(s.equalsIgnoreCase("UnknownEscaped")) {									// UnknownEscaped
			e.valid = false; }
		else if(s.equalsIgnoreCase("SequenzeNumber")) {									// Sequenze Number
			e.status = 0xff; e.dataVector.addElement(new Integer(0x00));
			v = new Integer(MIDIUtil.getNextValidToken(fpp)).intValue();
			e.dataVector.addElement(new Integer(v >> 8));
			e.dataVector.addElement(new Integer(v & 0xff)); }
		else if(s.equalsIgnoreCase("Text")) {														// Text
			e.status = 0xff; e.dataVector.addElement(new Integer(0x01));
			storeText(e,fpp); }
		else if(s.equalsIgnoreCase("Copyright")) {											// Copyright
			e.status = 0xff; e.dataVector.addElement(new Integer(0x02));
			storeText(e,fpp); }
		else if(s.equalsIgnoreCase("SequenzeName")) {										// Sequenze Name
			e.status = 0xff; e.dataVector.addElement(new Integer(0x03));
			storeText(e,fpp); }
		else if(s.equalsIgnoreCase("Instrument")) {											// Instrument
			e.status = 0xff; e.dataVector.addElement(new Integer(0x04));
			storeText(e,fpp); }
		else if(s.equalsIgnoreCase("Lyric")) {													// Lyric
			e.status = 0xff; e.dataVector.addElement(new Integer(0x05));
			storeText(e,fpp); }
		else if(s.equalsIgnoreCase("Marker")) {													// Marker
			e.status = 0xff; e.dataVector.addElement(new Integer(0x06));
			storeText(e,fpp); }
		else if(s.equalsIgnoreCase("CuePoint")) {												// Cue Point
			e.status = 0xff; e.dataVector.addElement(new Integer(0x07));
			storeText(e,fpp); }
		else if(s.equalsIgnoreCase("MIDI_Channel")) {										// MIDI Channel
			e.status = 0xff; e.dataVector.addElement(new Integer(0x20));
			i = new Integer(MIDIUtil.getNextValidToken(fpp));
			e.dataVector.addElement(i); }
		else if(s.equalsIgnoreCase("MIDI_Port")) {											// MIDI Port
			e.status = 0xff; e.dataVector.addElement(new Integer(0x21));
			i = new Integer(MIDIUtil.getNextValidToken(fpp));
			e.dataVector.addElement(i); }
		else if(s.equalsIgnoreCase("EndOfTrack")) {											// End of Track
			e.status = 0xff; e.dataVector.addElement(new Integer(0x2f)); }
		else if(s.equalsIgnoreCase("Tempo")) {													// Tempo
			e.status = 0xff; e.dataVector.addElement(new Integer(0x51));
			v = new Integer(MIDIUtil.getNextValidToken(fpp)).intValue();
			MIDIUtil.setTempo(v);
			e.dataVector.addElement(new Integer(v >> 16));
			e.dataVector.addElement(new Integer((v >> 8) & 0xff));
			e.dataVector.addElement(new Integer(v & 0xff)); }
		else if(s.equalsIgnoreCase("SMPTE_Offset")) {										// SMPTE Offset
			e.status = 0xff; e.dataVector.addElement(new Integer(0x54));
			String all = MIDIUtil.getNextValidToken(fpp);
			for(int c = 0; c < 5; c++) {
				i = new Integer(all.substring(c*3,c*3+2));
				e.dataVector.addElement(i);
			}}
		else if(s.equalsIgnoreCase("TimeSignature")) {									// Time Signature
			e.status = 0xff; e.dataVector.addElement(new Integer(0x58));
			String bar = MIDIUtil.getNextValidToken(fpp);
			for(int c = 0; c < bar.length(); c++) if(bar.charAt(c) == '/') v = c;
			i = new Integer(bar.substring(0,v));
			MIDIUtil.setNumerator(i.intValue());
			e.dataVector.addElement(i);
			v = new Integer(bar.substring(v+1,bar.length())).intValue();
			for(int c = 0; c < 8; c++) if((v >> c) == 1) i = new Integer(c);
			MIDIUtil.setDenominator(i.intValue());
			e.dataVector.addElement(i);
			i = new Integer(MIDIUtil.getNextValidToken(fpp));
			e.dataVector.addElement(i);
			i = new Integer(MIDIUtil.getNextValidToken(fpp));
			e.dataVector.addElement(i); }
		else if(s.equalsIgnoreCase("KeySignature")) {										// Key Signature
			e.status = 0xff; e.dataVector.addElement(new Integer(0x59));
			v = new Integer(MIDIUtil.getNextValidToken(fpp)).intValue();
			String sig = MIDIUtil.getNextValidToken(fpp);
			if(sig.equalsIgnoreCase("flat")) v = -v;
			e.dataVector.addElement(new Integer(v));
			String min = MIDIUtil.getNextValidToken(fpp);
			if(min.equalsIgnoreCase("minor")) v = 1;
			else if(min.equalsIgnoreCase("major")) v = 0;
			e.dataVector.addElement(new Integer(v)); }
		else if(s.equalsIgnoreCase("ProprietaryEvnt")) {								// Proprietary Evnt
			e.status = 0xff; e.dataVector.addElement(new Integer(0x7f));
			getHexValue(e,fpp); }
		else if(s.equalsIgnoreCase("UnknownMetaEvnt")) {								// Unknown Meta
			e.valid = false; }
		else { 																													// Unknown
			System.out.println("Unknown Keyword: \'"+s+"\'");
			e.valid = false; }
		if(!e.valid()) {
			MIDIUtil.skipLine(fpp);			// to skip all following dataTokens
			e.absTime -= e.deltaTime; }	// to ensure correct deltaTime of next Event
	}

}
