// MIDIConverter stores standard MIDI-files as ASCII-TXT-files and vice versa
// Copyright (C) 2000 Michael Bliem      email: e9271117@student.tuwien.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


import java.io.*;
import java.util.*;

import MIDIUtil;	// Utilities for handling IO, VarLenCode, MIDI-Spec...
import MIDIEvent;	// instance of this class stores one MIDI-event
import MIDIFrame;	// instance of window with textArea

public class MIDITrack
// Each Instance stores one MIDI-Track with all MIDIEvents
{
	int trackNumber;			// marks trackNumber
	boolean valid;				// marks track as valid track
	Vector eventVector;		// Vector for storing the MIDI-events

	public boolean valid() { return valid; }
	public int trackNumber() { return trackNumber; }
	public void setTrackNumber(int i) { trackNumber = i; }

	public MIDITrack(FileInputStream readFile) throws MIDIException
	// init-Method for an MIDITrack instance
	{
		super();
		trackNumber = 0;
		valid = true;
		eventVector = new Vector();
		if(MIDIUtil.FILETYPE) fromMIDI(readFile);
		else fromTXT(readFile);
	}

	void fromTXT(FileInputStream readFile) throws MIDIException
	// reads all MIDIEvents out of a TXT-file; count[0] for absTime; count[1] for end
	{
		String s = MIDIUtil.getNextValidToken(readFile);
		if(s.equals("EOF")) { valid = false; return; }
		trackNumber = new Integer(s).intValue();
		MIDIEvent event;
		int count[] = {0,0};
		int c = 0;	// counter for working output
		while(count[1] == 0) {
			event = new MIDIEvent(0,count,readFile);
			if((count[1] == 0) && event.valid()) eventVector.addElement(event);
			if((MIDIFrame.mText != null) && ((++c%250) == 0)) MIDIFrame.mText.append(".");
		}	// indicates working if Windows Version
	}

	void fromMIDI(FileInputStream readFile) throws MIDIException
	// reads all MIDIEvents out of a MIDI-file
	{
		try {
			MIDIEvent event;										// the actual MIDI-event
			int runningStatus = 0;							// remember Running Status
			if(!MIDIUtil.getString(4,readFile).equals("MTrk")) {
				readFile.skip(MIDIUtil.getInt(4,readFile));	// skips unknown chunks
				System.out.println("Unknown track ignored.\n");
				valid = false;
				return; }
			int trackSize = MIDIUtil.getInt(4,readFile);
			int count[] = {0,0};								// used to communicate with events
			while((trackSize - count[0]) > 0) {	// stores all MIDI-events
				event = new MIDIEvent(runningStatus,count,readFile);
				runningStatus = event.status;
				if(event.valid()) eventVector.addElement(event);
			}}
		catch (IOException m) { throw new MIDIException("Could not read MIDI-file.\n"); }
	}

	public void toTXT(FileOutputStream writeFile) throws MIDIException
	// writes all MIDIEvents to a TXT-file
	{
		MIDIUtil.putString("          Track ",writeFile);
		MIDIUtil.putString(new Integer(trackNumber).toString()+"\n",writeFile);
		Integer i = new Integer(MIDIUtil.BAROFFSET);
		MIDIUtil.putString("000"+ i.toString() + ":1:000\n",writeFile); // startTime
		int c = 0;	// counter for working output
		for(Enumeration e = eventVector.elements(); e.hasMoreElements();) {
			MIDIEvent event = (MIDIEvent)(e.nextElement());
			event.toTXT(trackNumber,writeFile);
			if((MIDIFrame.mText != null) && ((++c%400) == 0)) MIDIFrame.mText.append(".");
		}	// indicates working if Windows Version
	}

	public void toMIDI(FileOutputStream writeFile) throws MIDIException
	// writes all MIDEvents to a MIDI-file
	{
		int runningStatus = 0;
		MIDIUtil.putString("MTrk",4,writeFile);
		MIDIUtil.putInt(size() - 8,4,writeFile);
		for(Enumeration e = eventVector.elements(); e.hasMoreElements();) {
			MIDIEvent event = (MIDIEvent)(e.nextElement());
			event.toMIDI(runningStatus,writeFile);
			runningStatus = event.status;
		}
	}

	public int size() throws MIDIException
	// bytes needed to represent track in MIDI-file (with header)
	{
		int size = 8;																										// header-Bytes
		int runningStatus = 0;
		for(Enumeration  e = eventVector.elements(); e.hasMoreElements();) {
			MIDIEvent event = (MIDIEvent)(e.nextElement());								// event-Bytes
			size += event.size(runningStatus);
			runningStatus = event.status; }
		return size;
	}

	void insertEvent(MIDIEvent insert)
	// inserts Event insert into this track after deleting any same existing one
	{
		int i;
		int delta;
		boolean done = false;
    insert.deltaTime = insert.absTime;							// for empty track
		deleteEvent(insert,insert.status & 0x0f);				// delete same existing event
		for(i = 0;i < eventVector.size();i++) {
			MIDIEvent e = (MIDIEvent)eventVector.elementAt(i);
			delta = insert.absTime - e.absTime;
			if(delta < 0) {																// last 'parallel' event
				insert.deltaTime = e.deltaTime + delta;
				e.deltaTime = -delta;
				eventVector.insertElementAt(insert,i);
				done = true;
				return; }
			insert.deltaTime = delta; }											// if last event
		if(!done) eventVector.insertElementAt(insert,i);	// if last event or empty track
	}

	void deleteEvent(int absTime, int status, int chn, byte[] data)
	// delete all Events with absTime (ignore if > -1), status (ignor if > -1),
	// channel chn (all if -1) and data (if not null) (and ignore all dataBytes = -1)
	{
		MIDIEvent e = new MIDIEvent();
		e.absTime = absTime;
		e.status = status;
		if(data != null) {
			e.dataVector = new Vector();
			for(int i = 0;i < data.length;i++) e.dataVector.addElement(new Integer(data[i]));
		}
		deleteEvent(e,chn);
	}

	void deleteEvent(MIDIEvent compare,int chn)
	// deletes any MIDIEvent that is equal to Event compare at chn (all if -1)
	// note: to delete systemMessages you must specify chn (eg. f7: chn = 7)
	// returns how many events were deleted
	{
		boolean equal;
		for(int i = 0;i < eventVector.size();i++) {
			equal = true;								// after equaltests are done, and still true -> delete
			MIDIEvent event = (MIDIEvent)eventVector.elementAt(i);
			if(((compare.absTime != -1) && (compare.absTime != event.absTime))
			|| ((compare.status  != -1) && ((compare.status & 0xf0) != (event.status & 0xf0)))
			|| ((chn != -1) && (compare.status == -1) && (event.status >= 0xf0))
			|| ((chn != -1) && (chn != (event.status & 0x0f))))	equal = false;
			else if(compare.dataVector != null)
			  if(compare.dataVector.size() != event.dataVector.size()) equal = false;
				else for(int c = 0;c < event.dataVector.size();c++) {
					Integer compareData = (Integer)compare.dataVector.elementAt(c);
					Integer eventData = (Integer)event.dataVector.elementAt(c);
					if((compareData.intValue() != -1)
					&& (compareData.intValue() != eventData.intValue())) equal = false;
				}
			if(equal) {
				if((event.deltaTime > 0) && (eventVector.size() > (i+1))) {
					MIDIEvent nextEvent = (MIDIEvent)eventVector.elementAt(i+1);
					nextEvent.deltaTime += event.deltaTime; }
				eventVector.removeElementAt(i--);						// i-- because of one event less
			}
		}
	}

}
