// MIDIConverter stores standard MIDI-files as ASCII-TXT-files and vice versa
// Copyright (C) 2000 Michael Bliem      email: e9271117@student.tuwien.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


import java.io.*;

public class MIDIUtil
// Utilities for handling IO, VarLenCode, Timeing ...
{
	// ------------------------------ Global Values ------------------------------------

	static int NUMERATOR = 4;						// numerator for TimeSignature
	static int DENOMINATOR  = 2;				// denominator for TimeSignature in 2expDEN
	static int TEMPO = 500000;					// tempo in Micro Seconds per Quarter
	static int SMPTE = 1;								// Frames Per Second for SMPTE if > 1
	static int DIVISION = 384;					// Division of Quarter or Frame

	static boolean FILETYPE = false;		// true: from MIDI- false: from TXT-file
	static boolean COMMENTS = true;			// if true: comments to ASCII-text added
	static boolean GERMAN = false;			// if true: b will be written as h
	static int PROGRAMOFFSET = 1;				// if 0: program numbers starts with 0
	static int BAROFFSET = 0;						// if 0: first bar is named bar 0

	public static void setNumerator(int i) { NUMERATOR = i; }
	public static void setDenominator(int i) { DENOMINATOR = i; }
	public static void setTempo(int i) {	TEMPO = i;	}
	public static void setSmpte(int i) {	SMPTE = i;	}
	public static void setDivision(int i) {	DIVISION = i;	}
	public static void setComments(boolean b) {	COMMENTS = b;	}
	public static void setFileType(boolean b) {	FILETYPE = b;	}
	public static void setGerman(boolean b) {	GERMAN = b;	}

	public static void setProgramOffset(int i)
	{ if(i > 0) PROGRAMOFFSET = 1; else PROGRAMOFFSET = 0; }

	public static void setBarOffset(int i)
	{ if(i > 0) BAROFFSET = 1; else BAROFFSET = 0; }


	// ---------------------------- MIDI Timing Methods --------------------------------


	public static double getTicks()
	// returns number of Ticks (deltaTime) per Beat
	{
		double t;
		if(SMPTE == 1) t = (double)((DIVISION << 2) >> DENOMINATOR);
		else t=((double)TEMPO * (double)(((SMPTE*DIVISION)<< 2) >> DENOMINATOR)) / 1000000.0;
		return t;
	}

	public static int getAbsTime(String s) throws MIDIException
	// returns absTime in clocks calculated out of string s in PPQN
	{
		try {
			int bar = new Integer(s.substring(0,4)).intValue() - BAROFFSET;
			int beat = new Integer(s.substring(5,6)).intValue() - 1;
			if(beat >= NUMERATOR) beat = NUMERATOR - 1;
			int div = new Integer(s.substring(7,10)).intValue();
			if((SMPTE == 1) && (div >= DIVISION)) div = DIVISION - 1;
			else if((SMPTE > 1) && (div >= (DIVISION * SMPTE))) div = (DIVISION * SMPTE) - 1;
			int value = bar * (int)(getTicks() * (double)NUMERATOR);
			value += (int)(getTicks() * (double)beat);
			return value + div; }
		catch(Exception m) { throw new MIDIException("No valid TimeStamp.\n"); }
	}


	// ------------------------------ File IO Methods ----------------------------------


	public static int getInt(int length,FileInputStream fpp) throws MIDIException
	// returns integer of (max 4 bytes) length out of file fpp
	{
		int result = 0;
		try {
			for(int i=1; i<length; i++) {
				result += fpp.read();
				result <<= 8; }
			result += fpp.read(); }
		catch(IOException m) {
			throw new MIDIException("File error: could not read.\n");}
		return result;
	}

	public static void putInt(int value,int length,FileOutputStream fpp)
		throws MIDIException
	// writes value in length bytes to file fpp
	{
		byte[] byteBuffer = new byte[length];
		for(int i=0; i<length; i++) {
			byteBuffer[length-(1+i)] = (byte)(value & 0xff);
			value >>= 8; }
		try { fpp.write(byteBuffer); }
		catch(IOException m) { throw new MIDIException("File error: could not write.\n");}
	}

	public static int getVarLen(int count[],FileInputStream fpp) throws MIDIException
	// reads VariableLengthCode out of File fpp
	{
		int result = 0;
		int i = 0x80;
		try {
			while((i & 0x80) == 0x80) {
				i = fpp.read(); count[0]++;
				result <<= 7;
				result |= (i & 0x7f);
			}}
		catch(IOException m) {
			throw new MIDIException("File error: could not read.\n"); }
		if(result > 0xfffffff) throw new MIDIException("Illegal VariableLengthCode.\n");
		return result;
	}

	public static void putVarLen(int value,FileOutputStream fpp) throws MIDIException
	// writes VariableLengthCode to File fpp
	{
		if(value > 0xfffffff) throw new MIDIException("Illegal VariableLengthCode.\n");
		byte[] byteBuffer = new byte[4];
		int result = value;
		int count = 0;
		while(result >= 128) {
			byteBuffer[count++] = (byte)((result & 0x7f) | 0x80);
			result >>= 7; }
		byteBuffer[count++] = (byte)((result & 0x7f) | 0x80);
		byteBuffer[0] &= 0x7f;
		try { for(int i = count; i>0; i--) fpp.write(byteBuffer[i-1]); }
		catch(IOException m) { throw new MIDIException("File error: could not write.\n"); }
	}

  public static String getString(int length,FileInputStream fpp) throws MIDIException
	// returns a String of length out of file fpp
	{
		byte[] byteBuffer = new byte[length];
		try { fpp.read(byteBuffer); }
		catch(IOException m) {
			throw new MIDIException("File error: could not read.\n");}
		return new String(byteBuffer);
	}

	public static void putString(String s,int length,FileOutputStream fpp)
		throws MIDIException
	// writes String of length to file fpp
	{
		byte[] byteBuffer = s.getBytes();
		try { for(int i=0; i<length; i++) fpp.write(byteBuffer[i]); }
		catch(IOException m) { throw new MIDIException("File error: could not write.\n");}
	}

	public static void putString(String s,FileOutputStream fpp) throws MIDIException
	// writes String to file fpp
	{ putString(s,s.length(),fpp); }

	public static void putComment(String s,FileOutputStream fpp)
		throws MIDIException
	// writes comment s to file fpp
	{
		String d = new String("--------");
		String c = new String("//" + d+d+d+d + " " + s + " " + d+d+d );
		try { c = c.substring(0,59); }
		catch(Exception m) { throw new MIDIException("Wrong String index.\n"); }
		c = c.concat("\n");
		putString(c,fpp);
	}

	public static void putCopyright(FileOutputStream fpp) throws MIDIException
	// writes a short copyright message to file fpp
	{
		putString("//\n// created by MIDIConverter\n",fpp);
		putString("// (c)2000 Michael Bliem\n",fpp);
		putString("// e9271117@student.tuwien.ac.at\n//\n",fpp);
	}

	public static String getLine(FileInputStream fpp) throws MIDIException
	// reads a line out of file fpp and returns as String
	{
		String s = new String();
		byte byteBuffer[] = new byte[1];
		try {
			while(true) {
				byteBuffer[0] = (byte)fpp.read();
				if((s.length() == 0) && (byteBuffer[0] < 0)) return null;
				if(((char)byteBuffer[0] == '\n') || (byteBuffer[0] < 0)) break;
				s = s.concat(new String(byteBuffer));
			}}
		catch(IOException m) { throw new MIDIException("File error: could not read.\n"); }
		return s;
	}

	public static void skipLine(FileInputStream fpp) throws MIDIException
	// skips a line of file fpp
	{
		byte b = 0;
		try { while(((char)b != '\n') && (b > -1)) b = (byte)fpp.read(); }
		catch(IOException m) { throw new MIDIException("File error: could not read.\n"); }
	}

	public static String getNextToken(FileInputStream fpp) throws MIDIException
	// reads a token out of file fpp and returns as string
	{
		String s = new String();
		byte buf[] = {0};
		boolean start = false;
		boolean quoted = false;
		try {
			while(buf[0] >= 0) {
				buf[0] = (byte)fpp.read();
				if(buf[0] < 0) break;
				else if((char)buf[0] == '\"') quoted = !quoted;
				if(((char)buf[0] == '\n') || ((char)buf[0] == '\r')) { if(start) break; }
				else if((((char)buf[0] != ' ') && ((char)buf[0] != '\t')) || quoted) {
					if(!start) start = true;
					s = s.concat(new String(buf)); }
				else if(start) break; }
			if((buf[0] < 0) && (s.length() == 0)) s = s.concat("EOF"); }
		catch(IOException m) { throw new MIDIException("File error: could not read.\n"); }
		return s;
	}

	public static String getNextValidToken(FileInputStream fpp) throws MIDIException
	// reads a valid token out of file fpp and returns as string
	{
		String s = null;
		try {
			boolean valid = false;
			while(!valid) {
				s = getNextToken(fpp);
				if((s.length() > 1) && (s.startsWith("//"))) skipLine(fpp);
				else valid = true;
			}}
		catch(Exception m) { throw new MIDIException("String handling error.\n"); }
		return s;
	}

}